#include <windows.h>

// Defines how frequently the thread procedure ticks over internally.
// See thread procedure code below.
#define	INTERNAL_SLEEP_PERIOD_MILLISECS		100

// Function prototypes
extern "C" int __stdcall SetupTimedTicks(int WindowHandle, int PeriodInSeconds);
extern "C" int __stdcall RemoveTimer(int TimerHandle);
DWORD CALLBACK TimerThreadProc(LPVOID pThreadParam);
HMODULE GetCurrentModule();


// Structure which is used to pass the client's parameters to the thread procedure 
typedef struct clientrequestdef {
	int WindowHandle;
	int PeriodInSeconds;
} CLIENTREQUESTDEF;


// Global (i.e. per-process, i.e. per-MT4-instance) variable which records
// whether we've increased the refcount on the DLL to keep it in memory
// See the LoadLibrary() call in SetupTickTimer()
BOOL glbBumpedUpDllRefCount = false;


// Called by an EA, typically in init(), in order to set up timed simulated
// ticks to its window. In other words, guarantees a call to start() at 
// least as frequently as PeriodInSeconds. The parameters are the chart's 
// window handle as provided by WindowHandle(Symbol(), Period()) and the 
// frequency at which the EA wants simulated ticks. The return value
// is a handle which the EA should store for later passing to RemoveTickTimer() 
// during deinit(). (If the EA doesn't do this, the chart window will keep getting 
// ticks for ever.)

extern "C" int __stdcall SetupTickTimer(int WindowHandle, int PeriodInSeconds)
{
	// Need a window handle
	if (!WindowHandle) return 0;

	// The only problem with creating subsidiary threads in MT4 DLLs is that
	// the DLL can get unloaded by MT4 while a thread is still active. Fatal
	// crash. To avoid this scenario, we do a one-time-only increase 
	// of the DLL's refcount, forcing it to remain in memory until 
	// MT4 terminates
	if (!glbBumpedUpDllRefCount) {
		char cLibPath[MAX_PATH + 1];
		GetModuleFileNameA(GetCurrentModule(), cLibPath, MAX_PATH);
		HMODULE me = LoadLibraryA(cLibPath);
		glbBumpedUpDllRefCount = true;
	}

	// Create a block of memory containing the parameters, for passing to the thread procedure.
	// The thread is responsible for freeing this memory.
	CLIENTREQUESTDEF * pReq = new CLIENTREQUESTDEF;	
	pReq->WindowHandle = WindowHandle;
	pReq->PeriodInSeconds = PeriodInSeconds;
	
	// Create a thread to do the timed ticks. (What the EA caller gets back is a raw
	// operating system thread handle.)
	DWORD lThreadId;
	return (int)CreateThread(NULL, 0, TimerThreadProc, pReq, 0, &lThreadId);
}


// Called by an EA, in deinit(), in order to terminate the timed ticks which
// it set up earlier. Returns 1 if termination is successful, otherwise 0.
// The EA passes in the handle which it got from SetupTickTimer() - this is a 
// raw thread handle, and the DLL simply does a TerminateThread()

extern "C" int __stdcall RemoveTickTimer(int TimerHandle)
{
	if (!TimerHandle) return 1;

	if (TerminateThread((HANDLE)TimerHandle, 0)) {
		return 1;
	} else {
		return 0;
	}
}


// The thread procedure which handles sending simulated ticks to a chart window.
// The parameter is a CLIENTREQUESTDEF which contains the parameters for 
// the call to SetupTickTimer(). This function is responsible for freeing
// this memory.

DWORD CALLBACK TimerThreadProc(LPVOID pThreadParam)
{
	// Cast the parameter to a CLIENTREQUESTDEF; extract the
	// values; and free the memory
	CLIENTREQUESTDEF * pReq = (CLIENTREQUESTDEF *)pThreadParam;
	HWND hWindow = (HWND)pReq->WindowHandle;
	int PeriodInSeconds = pReq->PeriodInSeconds;
	delete pReq;

	// Do a one-off retrieval of the message number which needs to be sent
	// to the MT4 window to simulate ticks
	int MessageNumber = RegisterWindowMessageA("MetaTrader4_Internal_Message");
	
	// This thread procedure ticks over more frequently than the caller
	// actually asks for, so that MT4 doesn't end up in a position where 
	// it has to wait for this thread procedure to terminate when MT4
	// is shutting down. We keep a count of how often the caller wants ticks, 
	// and decrease it by the length of the internal sleep every time
	// we do an internal sleep.
	int MilliSecondsUntilNextPost = PeriodInSeconds * 1000;
	
	// Keep looping indefinitely until either (a) the RemoveTickTimer()
	// function is called, or (b) the window ceases to exist - handling
	// scenarios where the caller fails to use RemoveTickTimer()
	while (IsWindow(hWindow)) {
		// Decrease the count of the time until the next simulated tick
		MilliSecondsUntilNextPost -= INTERNAL_SLEEP_PERIOD_MILLISECS;

		// If it's time for the next simulated tick, carry it out
		// and reset the sleep timer
		if (MilliSecondsUntilNextPost <= 0) {
			PostMessageA(hWindow, MessageNumber, 2, 0);
			MilliSecondsUntilNextPost = PeriodInSeconds * 1000;
		}
		
		// Do a short internal sleep
		Sleep(INTERNAL_SLEEP_PERIOD_MILLISECS);
	}

	return (0);
}



// Service function for providing the HMODULE of a DLL. Required
// for the call to LoadLibrary() above. Age-old code from MSDN.

#if _MSC_VER >= 1300    // for VC 7.0
  // from ATL 7.0 sources
  #ifndef _delayimp_h
  extern "C" IMAGE_DOS_HEADER __ImageBase;
  #endif
#endif

HMODULE GetCurrentModule()
{
#if _MSC_VER < 1300    // earlier than .NET compiler (VC 6.0)

  // Here's a trick that will get you the handle of the module
  // you're running in without any a-priori knowledge

  MEMORY_BASIC_INFORMATION mbi;
  static int dummy;
  VirtualQuery( &dummy, &mbi, sizeof(mbi) );

  return reinterpret_cast<HMODULE>(mbi.AllocationBase);

#else    // VC 7.0

  // from ATL 7.0 sources

  return reinterpret_cast<HMODULE>(&__ImageBase);
#endif
}

